#include <v_lib.h>

/*
 * Binary tree image decomposition
 */

typedef struct
{
	u32_t depth;
	u32_t left, top, right, bottom;
	vec_t value;
} btdecomp_node_t, *btdecomp_node_p;

typedef struct btdecomp_leaf_info_s
{
	u32_t depth;
	u32_t left,top,right,bottom;
	u32_t area;
	vec_t value;
}	btdecomp_leaf_info_t, *btdecomp_leaf_info_p;

typedef struct bt_decomp_s
{
	btree_p		root;
	bstack_t 	info_stack;
	u32_t		source_rows, source_cols;
	u32_t		min_depth, max_depth, depth_levels; /* max - min + 1 */
	u32_t		min_area, max_area;
	vec_t		min_value, max_value;
	u32_p		leafs_at_depth; /* leafs_at_depth[level] == num_leafs at this level */
	vector_p	*values_at_depth;
	vector_p	*histogram_at_depth;
	vector_p	*histogram_space_at_depth;
}	bt_decomp_t, *bt_decomp_p;

V_LOCAL void
submatMinMax(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t *min_ret,
	vec_t *max_ret)
{
	u32_t r, c;
	vec_t min, max, *row;
	min = max = data->cell[top][left];
	for(r = top;r <= bottom;r++)
	{
		row = data->cell[r];
		for(c = left;c <= right;c++)
		{
			if(min > row[c])
				min = row[c];
			else if(max < row[c])
				max = row[c];
		}
	}
	if(min_ret)
		*min_ret = min;
	if(max_ret)
		*max_ret = max;
}

V_LOCAL btree_p V_CALL
bt_img_decomp_r(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t thresold,
	u32_t depth)
{
	u32_t vsplit, hsplit;
	vec_t min, max;
	btree_p node;
	btdecomp_node_t node_data;
	vboolean vspan = left == right; /* vertical span of pixels */
	vboolean hspan = top == bottom; /* horizontal span */
	node_data.left = left;
	node_data.top = top;
	node_data.right = right;
	node_data.bottom = bottom;
	node_data.depth = depth;
	if(hspan && vspan)
	{
		/* one pixel: leaf */
		node_data.value = data->cell[top][left];
		return btStore(&node_data,sizeof(node_data),NULL,0);
	}
	submatMinMax(data,left,top,right,bottom,&min,&max);
	node_data.value = (min + max) * (vec_t)0.5;
	node = btStore(&node_data,sizeof(node_data),NULL,0);
	if((max - min) < thresold) /* difference too small: leaf */
		return node;
	depth++;
	/* choose split direction */
	if((right-left) > (bottom-top))
	{
		/* split into left and right submatrices */
		hsplit = left + ((right - left) / 2);
		node->children[0] = bt_img_decomp_r(data,left,top,hsplit,bottom,thresold,depth);
		node->children[0]->parent = node;
		node->children[1] = bt_img_decomp_r(data,hsplit+1,top,right,bottom,thresold,depth);
		node->children[1]->parent = node;
	}
	else
	{
		/* split into upper and lower submatrices */
		vsplit = top + ((bottom - top) / 2);
		node->children[0] = bt_img_decomp_r(data,left,top,right,vsplit,thresold,depth);
		node->children[0]->parent = node;
		node->children[1] = bt_img_decomp_r(data,left,vsplit+1,right,bottom,thresold,depth);
		node->children[1]->parent = node;
	}
	return node;
}

V_LOCAL void
bt_img_describe_r(bstack_p info_stack,btree_p root)
{
	btdecomp_node_p nd;
	btdecomp_leaf_info_t info;
	if(!root->children[0]) /* leaf */
	{
		nd = (btdecomp_node_p)root->data;
		info.depth = nd->depth;
		info.left = nd->left;
		info.top = nd->top;
		info.right = nd->right;
		info.bottom = nd->bottom;
		info.area = (info.right - info.left + 1) * (info.bottom - info.top + 1);
		info.value = nd->value;
		stackPush(info_stack,&info);
		return;
	}
	bt_img_describe_r(info_stack,root->children[0]);
	bt_img_describe_r(info_stack,root->children[1]);
}

/*
 * vecHistogram()
 */
vboolean
vecHistogram(vector_p dest,vector_p src,vboolean normalize,
	vector_p space)
{
	u32_t sum, i, j, n_divs = dest->size;
    vec_t lower, upper, min, max, delta, *ptr;
    ptr = src->elements;
    min = ptr[vecMin(src)];
    max = ptr[vecMax(src)];
    delta = (max - min) / n_divs;
    V_ASSERT(!space || (space->size = n_divs));
    n_divs--;
    for(i = 0;i < n_divs;i++)
    {
    	sum = 0;
        lower = min + i * delta;
        upper = lower + delta;
    	if(space)
    		space->elements[i] = (lower + upper) / 2;
        for(j = 0;j < src->size;j++)
        {
            if(ptr[j] >= lower && ptr[j] < upper)
				sum++;
		}
        dest->elements[i] = (vec_t)sum;
    }
    /* last bin */
   	sum = 0;
	lower = min + i * delta;
	if(space)
		space->elements[i] = (2 * lower + delta) / 2;
	for(j = 0;j < src->size;j++)
	{
		if(ptr[j] >= lower)
			sum++;
	}
	dest->elements[i] = (vec_t)sum;
	n_divs++;
    if(normalize)
    {
    	vec_t vsum = dest->elements[0];
        for(i = 1;i < n_divs;i++)
        	vsum += dest->elements[i];
		if(vsum > 0)
        {
        	vsum = 1 / vsum;
            vecScale(dest,dest,vsum);
        }
        else
        	return false;
    }
    return true;
}

bt_decomp_p
iplDecomposeImage(matrix_p m,vec_t thresold,
	u32_t histogram_divs)
{
	u32_t i, d_offs, num_leafs, *counts;
	btdecomp_leaf_info_p base, ptr;
	bt_decomp_p self = (bt_decomp_p)memAllocZ(sizeof(bt_decomp_t));
	self->root = bt_img_decomp_r(m,0,0,m->cols - 1,m->rows - 1,thresold,0);
	self->source_rows = m->rows;
	self->source_cols = m->cols;
	stackInit(&self->info_stack,sizeof(btdecomp_leaf_info_t));
	bt_img_describe_r(&self->info_stack,self->root);
	num_leafs = stackCount(&self->info_stack);
	base = (btdecomp_leaf_info_p)stackGetStorage(&self->info_stack);
	self->min_depth = MAXLONG;
	self->max_depth = 0;
	self->min_area = m->rows * m->cols;
	self->max_area = 0;
	self->min_value = MAXFLOAT;
	self->max_value = -MAXFLOAT;
	for(i = 0;i < num_leafs;i++)
	{
		ptr = base + i;
		if(self->min_depth > ptr->depth)
			self->min_depth = ptr->depth;
		if(self->max_depth < ptr->depth)
			self->max_depth = ptr->depth;
		if(self->min_area > ptr->area)
			self->min_area = ptr->area;
		if(self->max_area < ptr->area)
			self->max_area = ptr->area;
		if(self->min_value > ptr->value)
			self->min_value = ptr->value;
		if(self->max_value < ptr->value)
			self->max_value = ptr->value;
	}
    self->depth_levels = self->max_depth - self->min_depth + 1;
    self->leafs_at_depth = (u32_p)memAllocZ(sizeof(u32_t) * self->depth_levels);
    counts = (u32_p)memAllocZ(sizeof(u32_t) * self->depth_levels);
	for(i = 0;i < num_leafs;i++)
	{
		ptr = base + i;
		self->leafs_at_depth[ptr->depth - self->min_depth]++;
	}
	self->values_at_depth =
		(vector_p*)memAlloc(sizeof(vector_p) * self->depth_levels);
	self->histogram_at_depth =
		(vector_p*)memAlloc(sizeof(vector_p) * self->depth_levels);
	self->histogram_space_at_depth =
		(vector_p*)memAlloc(sizeof(vector_p) * self->depth_levels);
	for(i = 0;i < self->depth_levels;i++)
	{
		if(self->leafs_at_depth[i])
			self->values_at_depth[i] = vecAlloc(self->leafs_at_depth[i]);
		else
			self->values_at_depth[i] = 0;
	}
	for(i = 0;i < num_leafs;i++)
	{
		ptr = base + i;
		d_offs = ptr->depth - self->min_depth;
		self->values_at_depth[d_offs]->elements[counts[d_offs]] = ptr->value;
		counts[d_offs]++;
	}
	for(i = 0;i < self->depth_levels;i++)
	{
		if(self->values_at_depth[i])
		{
			self->histogram_at_depth[i] = vecAlloc(histogram_divs);
			self->histogram_space_at_depth[i] = vecAlloc(histogram_divs);
			vecHistogram(self->histogram_at_depth[i],self->values_at_depth[i],
				false,self->histogram_space_at_depth[i]);
		}
		else
		{
			self->histogram_at_depth[i] = NULL;
			self->histogram_space_at_depth[i] = NULL;
		}
	}
	memFree(counts);
	return self;
}
	
void
iplSaveDecompositionDescription(char *fname,
	bt_decomp_p decomp,
	vboolean save_values,vboolean save_histograms,
	vboolean histogram_save_all,
	vboolean relative_frequencies)
{
	u32_t i;
	vec_p ptr, sptr, endptr;
	vec_t freq_scale = 1;
	vhandle_t out = fileOpenWrite(fname);
	V_ASSERT(out != LIBV_INVALID_HANDLE);
	V_ASSERT(decomp);
	filePrintf(out,"source_dimension (%u,%u)\n",
		decomp->source_rows,decomp->source_cols);
    filePrintf(out,"depth_range (%u,%u)\n",
    	decomp->min_depth,decomp->max_depth);
    filePrintf(out,"area_range (%u,%u)\n",
    	decomp->min_area,decomp->max_area);
    filePrintf(out,"value_range (%g,%g)\n",
    	decomp->min_value,decomp->max_value);
	if(!save_values && !save_histograms)
	{
		for(i = 0;i < decomp->depth_levels;i++)
			filePrintf(out,"%u\t%u\n",i + decomp->min_depth,
				decomp->leafs_at_depth[i]);
	    fileClose(out);
	    return;
	}
	for(i = 0;i < decomp->depth_levels;i++)
	{
		filePrintf(out,"at_depth %u\n",i + decomp->min_depth);
		filePrintf(out,"leafs %u\n",decomp->leafs_at_depth[i]);
		if(save_values)
		{
			filePrintf(out,"values\n");
			if(decomp->values_at_depth[i])
			{
				ptr = decomp->values_at_depth[i]->elements;
				endptr = ptr + decomp->values_at_depth[i]->size;
				for(;ptr < endptr;ptr++)
					filePrintf(out,"%g\n",*ptr);
			}
		}
		if(save_histograms)
		{
			if(relative_frequencies && decomp->values_at_depth[i])
			{
				freq_scale = (vec_t)1.0 / decomp->leafs_at_depth[i];
			}
			else
				freq_scale = 1;
			if(decomp->histogram_at_depth[i])
			{
				filePrintf(out,"histogram\n");
				ptr = decomp->histogram_at_depth[i]->elements;
				sptr = decomp->histogram_space_at_depth[i]->elements;
				endptr = ptr + decomp->histogram_at_depth[i]->size;
				if(histogram_save_all)
				{
					for(;ptr < endptr;ptr++,sptr++)
						filePrintf(out,"%g\t%g\n",*sptr,*ptr * freq_scale);
				}
				else
				{
					for(;ptr < endptr;ptr++,sptr++)
						if(!IsZero(*ptr))
							filePrintf(out,"%g\t%g\n",*sptr,*ptr);
				}
			}
		}
	}
	fileClose(out);
}

void
iplFreeDecomposition(bt_decomp_p decomp)
{
	u32_t i;
	if(!decomp)
		return;
	btDeleteAllNodes(decomp->root,NULL);
	stackClear(&decomp->info_stack);
	for(i = 0;i < decomp->depth_levels;i++)
	{
		if(decomp->values_at_depth[i])
			memFree(decomp->values_at_depth[i]);
		if(decomp->histogram_at_depth[i])
			memFree(decomp->histogram_at_depth[i]);
		if(decomp->histogram_space_at_depth[i])
			memFree(decomp->histogram_space_at_depth[i]);
	}
	memFree(decomp->values_at_depth);
	memFree(decomp->histogram_at_depth);
	memFree(decomp->histogram_space_at_depth);
	memFree(decomp->leafs_at_depth);
	memFree(decomp);
}

V_LOCAL void
matSetRect(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t value)
{
	u32_t r, c;
	vec_t *row;
	for(r = top;r <= bottom;r++)
	{
		row = data->cell[r];
		for(c = left;c <= right;c++)
			row[c] = value;
	}
}

V_LOCAL void
bt_img_draw_decomp_r(matrix_p surface,btree_p root)
{
	btdecomp_node_p nd;
	if(!root->children[0]) /* leaf */
	{
		nd = (btdecomp_node_p)root->data;
		matSetRect(surface,nd->left,nd->top,nd->right,nd->bottom,nd->value);
		return;
	}
	bt_img_draw_decomp_r(surface,root->children[0]);
	bt_img_draw_decomp_r(surface,root->children[1]);
}

V_LOCAL void
bt_img_draw_scale_decomp_r(matrix_p surface,btree_p root,
	u32_t depth_base,vec_t depth_to_color[])
{
	btdecomp_node_p nd;
	if(!root->children[0]) /* leaf */
	{
		nd = (btdecomp_node_p)root->data;
		matSetRect(surface,nd->left,nd->top,nd->right,nd->bottom,
			depth_to_color[nd->depth - depth_base]);
		return;
	}
	bt_img_draw_scale_decomp_r(surface,root->children[0],depth_base,depth_to_color);
	bt_img_draw_scale_decomp_r(surface,root->children[1],depth_base,depth_to_color);
}

void
iplDrawDecomposition(matrix_p surface,bt_decomp_p decomp)
{
	V_ASSERT(surface->rows == decomp->source_rows);
	V_ASSERT(surface->cols == decomp->source_cols);
	bt_img_draw_decomp_r(surface,decomp->root);
}

void
iplDrawDecompositionScale(matrix_p surface,bt_decomp_p decomp,vector_p scale)
{
	V_ASSERT(surface->rows == decomp->source_rows);
	V_ASSERT(surface->cols == decomp->source_cols);
	V_ASSERT(scale->size == decomp->depth_levels);
	bt_img_draw_scale_decomp_r(surface,decomp->root,decomp->min_depth,scale->elements);
}

void
iplDrawDecompositionLinearScale(matrix_p surface,bt_decomp_p decomp,
	vec_t min_value,vec_t max_value)
{
	u32_t i, num_colors = decomp->depth_levels;
	vec_t value, dv;
	vector_p scale;
	V_ASSERT(surface->rows == decomp->source_rows);
	V_ASSERT(surface->cols == decomp->source_cols);
	scale = vecAlloc(num_colors);
	scale->elements[0] = min_value;
	/* FIXME: vecLinSpaced(scale,min_value,max_value); */
	if(num_colors > 1)
	{
		dv = (max_value - min_value) / (num_colors - 1);
		value = min_value + dv;
		for(i = 1;i < num_colors;i++,value += dv)
			scale->elements[i] = value;
	}
	bt_img_draw_scale_decomp_r(surface,decomp->root,decomp->min_depth,scale->elements);
	memFree(scale);
}

/*
 * ===============================================================
 */

matrix_p load_image(char *name)
{
	u32_t w,h,i,j;
	matrix_p m;
	vec_t scale = 1.0 / 3 / 256;
	u8_t *ptr,*img = jpgLoad(name,&w,&h);
	if(!img)
		return NULL;
	m = matAlloc(w,h);
	ptr = img;
	for(i=0;i<m->rows;i++)
		for(j=0;j<m->cols;j++)
		{
			m->cell[i][j] = (ptr[0]+ptr[1]+ptr[2]) * scale;
			ptr += 4;
		}
	memFree(img);
	return m;
}

void save_image(char *name,matrix_p src)
{
	u32_t i,j;
	vec_t val;
	u8_t *ptr, *img = (u8_p)memAllocZ(src->rows * src->cols * 4);
	if(!img)
		return;
	ptr = img;
	for(i=0;i < src->rows;i++)
		for(j=0;j < src->cols;j++)
		{
			val = src->cell[i][j] * 255;
			if(val < 0)
				val = 0;
			else if(val > 255)
				val = 255;
			ptr[0] = ptr[1] = ptr[2] = (u8_t)val;
			ptr += 4;
		}
	bmpSave(name,src->cols,src->rows,img);
	memFree(img);
}

int main()
{
	bt_decomp_p decomp;
	matrix_p img;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	img = load_image("container.jpg");
	assert(img);
	decomp = iplDecomposeImage(img,0.3,256);
	iplDrawDecompositionLinearScale(img,decomp,0,1);
	iplFreeDecomposition(decomp);
	save_image("out1.bmp",img);
    memFree(img);
	memPrintStats();
	return 0;
}
