/*
 * Cut - and - paste from 'Numerical recipies in C'.
 */
#define NRC_FFT_SWAP(a, b) tempr = (a); (a) = (b); (b) = tempr

/*
 * Replaces data by its discrete Fourier transform, if isign is input as 1;
 * or replaces data by nn times its inverse discrete Fourier transform,
 * if isign is input as -1.
 * data is a complex array of length nn, input as a real array data[1..2*nn].
 * nn MUST be an integer power of 2.
 */

V_LOCAL void
nrc_four1(vec_t data[], int nn, int isign)
{
	int i, n, mmax, m, j, istep;
	vec_t wtemp, wr, wpr, wpi, wi, theta;
	vec_t tempr, tempi;
	n = nn << 1;
	j = 1;
	/* This is the bit-reversal section of the routine. */
	for (i = 1; i < n; i += 2)
	{
    	if (j > i)
		{
			/* Exchange the two complex numbers. */
	  		NRC_FFT_SWAP(data[j  ],data[i  ]);
			NRC_FFT_SWAP(data[j+1],data[i+1]);
		};
		m = n >> 1;
		while (m >= 2 && j > m)
		{
			j  -= m;
			m >>= 1;
		};
		j += m;
	};
	/* Here begins the Danielson-Lanczos section of the routine. */
	mmax = 2;
	while (n > mmax)
	{
		/* Initialize trigonometric recurrence. */
		theta	= (vec_t)6.28318530717959/(isign*mmax);
		wtemp	= (vec_t)sin(0.5*theta);
		wpr	= (vec_t)(-2.0)*wtemp*wtemp;
		wpi	= (vec_t)sin(theta);
		wr	= 1;
		wi	= 0;
		/* Here are the two nested inner loops. */
		istep	= 2*mmax;
		for (m = 1; m < mmax; m += 2)
		{
			for (i = m; i <= n; i += istep)
			{
				/* This is the Danielson-Lanczos formula. */
				j = i + mmax;
				tempr = wr*data[j  ] - wi*data[j+1];
				tempi = wr*data[j+1] + wi*data[j  ];
				data[j  ] = data[i  ] - tempr;
				data[j+1] = data[i+1] - tempi;
				data[i  ] += tempr;
				data[i+1] += tempi;
			}; /* Trigonometric recurrence. */
			wr = (wtemp = wr)*wpr - wi*wpi + wr;
			wi = wi*wpr + wtemp*wpi + wi;
		};
		mmax = istep;
	};
}

/*
 * Again, cut-and-paste
 */

/*
  Calculates the Fourier Transform of a set of 2n real-valued data points
  stored in array data[1..2n] replacing it by the positive frequency half
  of it's complex Fourier Transform if isign is set to +1.
  The real-valued first and last components of the complex transform
  are returned as elements data[1] and data[2] respectively.
  n must be a power of 2.
  Calculates the inverse transform of a complex data array
  which is the transform of real data if isign is set to -1.
  But the inverse transform array should be multiplied by 1/n.
*/

V_LOCAL void
nrc_realft(vec_t data[],int n,int isign)
{
	int	 i, i1, i2, i3, i4, n2p3;
	vec_t c1 = 0.5, c2, h1r, h1i, h2r, h2i;
	vec_t wr, wi, wpr, wpi, wtemp, theta;
	theta = (vec_t)M_PI/(vec_t)n;
	if(isign == 1)
	{
		c2 = -(vec_t)0.5;
		nrc_four1(data, n, 1);
	}
	else
	{
		c2 = (vec_t)0.5;
		theta = -theta;
	}
	wtemp = (vec_t)sin(0.5*theta);
	wpr = -(vec_t)2.0 * wtemp * wtemp;
	wpi = (vec_t)sin(theta);
	wr = 1 + wpr;
	wi = wpi;
	n2p3 = 2*n + 3;
	for (i = 2; i <= n/2; i++ )
	{
		i4	 =   1 + (i3 = n2p3 - (i2 = 1 + (i1 = i + i - 1)));
		h1r	 =  c1*(data[i1] + data[i3]);
		h1i	 =  c1*(data[i2] - data[i4]);
		h2r	 = -c2*(data[i2] + data[i4]);
		h2i	 =  c2*(data[i1] - data[i3]);
		data[i1]	 =  h1r + wr*h2r - wi*h2i;
		data[i2]	 =  h1i + wr*h2i + wi*h2r;
		data[i3]	 =  h1r - wr*h2r + wi*h2i;
		data[i4]	 = -h1i + wr*h2i + wi*h2r;
		wr	 = (wtemp = wr) * wpr - wi * wpi + wr;
		wi	 =  wi*wpr + wtemp*wpi + wi;
	}
	if (isign == 1)
    {
		data[1]	 = (h1r = data[1]) + data[2];
		data[2]	 =  h1r - data[2];
	}
	else
    {
		data[1]	 =  c1*((h1r = data[1]) + data[2]);
		data[2]	 =  c1*(h1r - data[2]);
		nrc_four1(data, n, -1);
    }
}

/*
 * Wrappers
 */

V_LOCAL void
fftComplexPacked(vec_t *packed,unsigned size_power,vboolean dir_sign)
{
	unsigned i,len;
	vec_t scale;
	len = 1 << size_power;
	if(dir_sign)
		nrc_four1(packed - 1,len >> 1,1);
	else
	{
		nrc_four1(packed - 1,len >> 1,-1);
		scale = (vec_t)2.0 / (vec_t)len;
		for(i=0;i < len;i++)
			packed[i] *= scale;
	}
}

V_LOCAL void
fftReal(vec_p samples,unsigned size_power,vboolean forward)
{
	vec_t scale;
	unsigned i, len;
	if(!samples||!size_power)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fftRealNip(): incorrect parameters.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	len = 1 << size_power;
	nrc_realft(samples - 1,len >> 1,forward ? 1 : -1);
	if(!forward)
	{
		scale = (vec_t)2.0 / (vec_t)len;
		for(i=0;i<len;i++)
			samples[i] *= scale;
	}
}	

