#include <v_lib.h>

#define	STR_EQ(s1,s2)		(!(strcasecmp(s1,s2)))

#define	CELL_OFFS			(EXIT_CODE_OFFSET + 1)
#define	ROW_OFFS			(EXIT_CODE_OFFSET + 2)
#define	COL_OFFS			(EXIT_CODE_OFFSET + 3)
#define	MAX_ROW_OFFS		(EXIT_CODE_OFFSET + 4)
#define	MAX_COL_OFFS		(EXIT_CODE_OFFSET + 5)
#define	RE_OFFS				(EXIT_CODE_OFFSET + 6)
#define	IM_OFFS				(EXIT_CODE_OFFSET + 7)
#define	MAG_OFFS			(EXIT_CODE_OFFSET + 8)
#define	PHASE_OFFS			(EXIT_CODE_OFFSET + 9)
#define	NEG_FREQ_OFFS		(EXIT_CODE_OFFSET + 10)
#define	ARG0_OFFS			(EXIT_CODE_OFFSET + 11)

#define	IPL_MAX_ARGS		8

V_LOCAL nsl_var_entry_t ipl_builtin_vars[] =
{
	{"$CELL",CELL_OFFS,"value of current cell (valid in 'matrix_transform' and 'vector_transform')"},
	{"$ROW",ROW_OFFS,"current row (valid in 'matrix_transform' and 'vector_transform')"},
	{"$COL",COL_OFFS,"current column (valid in 'matrix_transform' only))"},
	{"$MAX_ROW",MAX_ROW_OFFS,"maximal index of row (num_rows - 1)"},
	{"$MAX_COL",MAX_COL_OFFS,"maximal index of column (num_cols - 1)"},
	{"$RE",RE_OFFS,"real part of current FFT cell (valid in 'cart_trans')"},
	{"$IM",IM_OFFS,"imaginary part of current FFT cell (valid in 'cart_trans')"},
	{"$MAG",MAG_OFFS,"magnitude of current FFT cell (valid in 'polar_trans')"},
	{"$PHASE",PHASE_OFFS,"phase of current FFT cell (valid in 'polar_trans')"},
	{"$NEG_FREQ",NEG_FREQ_OFFS,"value of one means negative spatial frequency (valid in 'polar_trans' and 'cart_trans')"},
	{"$ARG0",ARG0_OFFS,"alias for argument #0"},
	{"$ARG1",ARG0_OFFS+1,"alias for argument #1"},
	{"$ARG2",ARG0_OFFS+2,"alias for argument #2"},
	{"$ARG3",ARG0_OFFS+3,"alias for argument #3"},
	{"$ARG4",ARG0_OFFS+4,"alias for argument #4"},
	{"$ARG5",ARG0_OFFS+5,"alias for argument #5"},
	{"$ARG6",ARG0_OFFS+6,"alias for argument #6"},
	{"$ARG7",ARG0_OFFS+7,"alias for argument #7"},
	{NULL,0,NULL}
};

V_LOCAL nsl_const_entry_t ipl_builtin_consts[] =
{
	{"Pi",M_PI,"3.14159265358979312..."},
	{"E",(vec_t)2.71828182845904509,"e = 2.71828182845904509..."},
	{NULL,0,NULL}
};

void save_image(char *name,matrix_p src)
{
	u32_t i,j;
	vec_t val;
	u8_t *ptr, *img = (u8_p)memAllocZ(src->rows * src->cols * 4);
	if(!img)
		return;
	ptr = img;
	for(i=0;i < src->rows;i++)
		for(j=0;j < src->cols;j++)
		{
			val = src->cell[i][j] * 255;
			if(val < 0)
				val = 0;
			else if(val > 255)
				val = 255;
			ptr[0] = ptr[1] = ptr[2] = (u8_t)val;
			ptr += 4;
		}
	bmpSave(name,src->cols,src->rows,img);
	memFree(img);
}

typedef struct ipl_transform_s
{
	char_p		name;
	vhandle_t	nsl_group;
	u32_t		num_args;
	vec_p		var_pool;
	vec_p		argp; /* &var_pool[ARG0_OFFS] */
	vp_stack_p	fn_init;
	vp_stack_p	fn_main;
}	ipl_transform_t, *ipl_transform_p;

void
iplFreeTransform(ipl_transform_p self)
{
	if(!self)
		return;
	if(self->name)
		memFree(self->name);
	if(self->nsl_group != LIBV_INVALID_HANDLE)
		nslDestroyFunctionGroup(self->nsl_group);
	memFree(self);
}

ipl_transform_p
iplLoadTransform(vhandle_t lexer)
{
	CALLER_NAME(iplLoadTransform)
	char *ptr, buf[16];
	vboolean arg_expected = true;
	vboolean has_init;
	vp_stack_t arg_aliases = TYPED_STACK_EMPTY_INITIALIZER;
	ipl_transform_p self = (ipl_transform_p)memAllocZ(sizeof(ipl_transform_t));
	self->nsl_group = LIBV_INVALID_HANDLE;
	/* assume keyword 'transform' already read in */
	lexPush(lexer,LEX_STATE_BIT);
	lexEnable(lexer,LEX_ALIAS_TRANSLATION);
	if(!lexNext(lexer))
	{
		SYN_UNEXPECTED_EOF(lexer);
		goto err_exit;
	}
	self->name = strDup(lexLast(lexer));
	if(!lexNextIs(lexer,"("))
	{
		SYN_EXPECTED(lexer,"(");
		goto err_exit;
	}
	/* parse argument list, create aliases */
	for(;;)
	{
		ptr = lexNext(lexer);
		if(!ptr)
		{
			SYN_UNEXPECTED_EOF(lexer);
			goto err_exit;
		}
		if(STR_EQ(ptr,"void"))
		{
			if(vpsCount(&arg_aliases))
			{
				SYN_UNEXPECTED(lexer,"void");
				goto err_exit;
			}
			if(!lexNextIs(lexer,")"))
			{
				SYN_EXPECTED(lexer,")");
				goto err_exit;
			}
		}
		if(STR_EQ(ptr,")"))
			break;
		if(STR_EQ(ptr,","))
		{
			if(arg_expected)
			{
				SYN_UNEXPECTED(lexer,",");
				goto err_exit;
			}
			arg_expected = true;
			continue;
		}
		arg_expected = false;
		if(vpsCount(&arg_aliases) == IPL_MAX_ARGS)
		{
			PARSE_ERROR(lexer,"too many arguments");
			goto err_exit;
		}
		ptr = strDup(ptr);
		sprintf(buf,"$ARG%u",vpsCount(&arg_aliases));
		vpsPush(&arg_aliases,ptr);
		lexDefineAlias(lexer,ptr,buf);
	}
	/* parse 'init' section */
	self->nsl_group = nslCreateFunctionGroup(ipl_builtin_vars,ipl_builtin_consts);
	V_ASSERT(self->nsl_group != LIBV_INVALID_HANDLE);
	if(!lexNextIs(lexer,"{"))
	{
		SYN_EXPECTED(lexer,"{");
		goto err_exit;
	}
	ptr = lexNext(lexer);
	if(ptr && STR_EQ(ptr,"init"))
	{
		if(!nslParseFunction(self->nsl_group,lexer))
			goto err_exit;
		has_init = true;
	}
	else
	{
		if(!ptr || !STR_EQ(ptr,"noinit"))
		{
			SYN_EXPECTED(lexer,"init|noinit");
			goto err_exit;
		}
		has_init = false;
	}
	/* parse 'main' section */
	ptr = lexNext(lexer);
	if(!ptr && !STR_EQ(ptr,"main"))
	{
		SYN_EXPECTED(lexer,"main");
		goto err_exit;
	}
	if(!nslParseFunction(self->nsl_group,lexer))
		goto err_exit;
	V_ASSERT(nslCreateExecutables(self->nsl_group));
	if(has_init)
	{
		V_ASSERT(nslGetExecutable(self->nsl_group,0,&self->fn_init,&self->var_pool));
		V_ASSERT(nslGetExecutable(self->nsl_group,1,&self->fn_main,NULL));
	}
	else
	{
		self->fn_init = NULL;
		V_ASSERT(nslGetExecutable(self->nsl_group,0,&self->fn_main,&self->var_pool));
	}
	self->argp = self->var_pool + ARG0_OFFS;
	if(!lexNextIs(lexer,"}"))
	{
		SYN_EXPECTED(lexer,"}");
		goto err_exit;
	}

	goto normal_exit;
err_exit:;
	iplFreeTransform(self);
	self = NULL;
normal_exit:;
	while(vpsPop(&arg_aliases,(void_p*)&ptr))
	{
		lexDefineAlias(lexer,ptr,NULL);
		memFree(ptr);
	}
	vpsClear(&arg_aliases);
	lexPop(lexer,LEX_STATE_BIT);
	return self;
}

V_LOCAL void
ipl_clip_rect(matrix_p m,
	u32_t *left,u32_t *top,u32_t *right,u32_t *bottom,
	vboolean frequency_domain)
{
	u32_t cols, rows;
	if(frequency_domain)
	{
		cols = m->cols >> 1;
		rows = m->rows >> 1;
	}
	else
	{
		cols = m->cols;
		rows = m->rows;
	}
	if(*right >= cols)
		*right = cols - 1;
	if(*bottom >= rows)
		*bottom = rows - 1;
	if(*left > *right)
		*left = *right;
	if(*top > *bottom)
		*top = *bottom;
}

void
iplExecTransform(matrix_p m,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	ipl_transform_p trans)
{
	u32_t i, j;
	vec_p vars = trans->var_pool;
	vars[MAX_ROW_OFFS] = (vec_t)(m->rows - 1);
	vars[MAX_COL_OFFS] = (vec_t)(m->cols - 1);
	if(trans->fn_init)
		NSL_EXECUTE(trans->fn_init,vars);
	ipl_clip_rect(m,&left,&top,&right,&bottom,false);
	for(i = top;i <= bottom;i++)
	{
		vars[ROW_OFFS] = (vec_t)i;
		for(j = left;j <= right;j++)
		{
			vars[COL_OFFS] = (vec_t)j;
			vars[CELL_OFFS] = m->cell[i][j];
			NSL_EXECUTE(trans->fn_main,vars);
			m->cell[i][j] = vars[CELL_OFFS];
		}
	}
}

V_LOCAL	vec_t
cplx_phase(vec_t re,vec_t im)
{
	/* reject degenerate cases */
	if(IsZero(re))
	{
		if(im < 0)
			return -M_PI2;
		return M_PI2;
	}
	return atan2(im, re);
}

void
iplExecCartTrans(matrix_p rvmft,
	u32_t col_start,u32_t row_start,
	u32_t col_end,u32_t row_end,
	ipl_transform_p trans)
{
	vec_p vars = trans->var_pool;
	u32_t r, nc, ftc, ftr;
	V_ASSERT(rvmft->cols > 1);
	V_ASSERT(IS_POWER_TWO(rvmft->cols) && IS_POWER_TWO(rvmft->rows));
	V_ASSERT(row_start <= row_end);
	V_ASSERT(col_start <= col_end);
	/* Clamp the ranges to positive part of Fourier transform */
	ipl_clip_rect(rvmft,&col_start,&row_start,&col_end,&row_end,true);
	vars[MAX_ROW_OFFS] = (vec_t)(rvmft->rows >> 1);
	vars[MAX_COL_OFFS] = (vec_t)((rvmft->cols >> 1) - 1);
	if(trans->fn_init)
		NSL_EXECUTE(trans->fn_init,vars);
	if(!row_start)
	{
		/* at zero frequency */
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		vars[ROW_OFFS] = 0;
		vars[NEG_FREQ_OFFS] = 0;
		for(;nc--;ftc += 2)
		{
			vars[COL_OFFS] = (vec_t)(ftc >> 1);
			vars[RE_OFFS] = rvmft->cell[0][ftc];
			vars[IM_OFFS] = rvmft->cell[0][ftc+1];
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[0][ftc] = vars[RE_OFFS];
			rvmft->cell[0][ftc+1] = vars[IM_OFFS];
		}
		row_start++;
	}
	/* at both positive and negative frequencies */
	for(r = row_start;r <= row_end;r++)
	{
		ftr = rvmft->rows - r;
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		vars[ROW_OFFS] = (vec_t)r;
		for(;nc--;ftc += 2)
		{
			vars[COL_OFFS] = (vec_t)(ftc >> 1);
			vars[RE_OFFS] = rvmft->cell[r][ftc];
			vars[IM_OFFS] = rvmft->cell[r][ftc+1];
			vars[NEG_FREQ_OFFS] = 0;
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[r][ftc] = vars[RE_OFFS];
			rvmft->cell[r][ftc+1] = vars[IM_OFFS];
			vars[RE_OFFS] = rvmft->cell[ftr][ftc];
			vars[IM_OFFS] = rvmft->cell[ftr][ftc+1];
			vars[NEG_FREQ_OFFS] = 1;
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[ftr][ftc] = vars[RE_OFFS];
			rvmft->cell[ftr][ftc+1] = vars[IM_OFFS];
		}
	}
	/* at highest frequency */
	if(r == (rvmft->rows >> 1))
	{
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		vars[ROW_OFFS] = (vec_t)r;
		for(;nc--;ftc += 2)
		{
			vars[COL_OFFS] = (vec_t)(ftc >> 1);
			vars[RE_OFFS] = rvmft->cell[r][ftc];
			vars[IM_OFFS] = rvmft->cell[r][ftc+1];
			vars[NEG_FREQ_OFFS] = 1;
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[r][ftc] = vars[RE_OFFS];
			rvmft->cell[r][ftc+1] = vars[IM_OFFS];
		}
	}
}

void
iplExecPolarTrans(matrix_p rvmft,
	u32_t col_start,u32_t row_start,
	u32_t col_end,u32_t row_end,
	ipl_transform_p trans)
{
	vec_p vars = trans->var_pool;
	vec_t re, im;
	u32_t r, nc, ftc, ftr;
	V_ASSERT(rvmft->cols > 1);
	V_ASSERT(IS_POWER_TWO(rvmft->cols) && IS_POWER_TWO(rvmft->rows));
	V_ASSERT(row_start <= row_end);
	V_ASSERT(col_start <= col_end);
	/* Clamp the ranges to positive part of Fourier transform */
	ipl_clip_rect(rvmft,&col_start,&row_start,&col_end,&row_end,true);
	vars[MAX_ROW_OFFS] = (vec_t)(rvmft->rows >> 1);
	vars[MAX_COL_OFFS] = (vec_t)((rvmft->cols >> 1) - 1);
	if(trans->fn_init)
		NSL_EXECUTE(trans->fn_init,vars);
	if(!row_start)
	{
		/* at zero frequency */
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		vars[ROW_OFFS] = 0;
		vars[NEG_FREQ_OFFS] = 0;
		while(nc--)
		{
			vars[COL_OFFS] = (vec_t)(ftc >> 1);
			re = rvmft->cell[0][ftc];
			im = rvmft->cell[0][ftc+1];
			vars[MAG_OFFS] = (vec_t)hypot(re, im);
			vars[PHASE_OFFS] = cplx_phase(re, im);
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[0][ftc++] = vars[MAG_OFFS] * (vec_t)cos(vars[PHASE_OFFS]);
			rvmft->cell[0][ftc++] = vars[MAG_OFFS] * (vec_t)sin(vars[PHASE_OFFS]);
		}
		row_start++;
	}
	/* at both positive and negative frequencies */
	for(r = row_start;r <= row_end;r++)
	{
		ftr = rvmft->rows - r;
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		vars[ROW_OFFS] = (vec_t)r;
		for(;nc--;ftc += 2)
		{
			vars[COL_OFFS] = (vec_t)(ftc >> 1);
			re = rvmft->cell[r][ftc];
			im = rvmft->cell[r][ftc+1];
			vars[MAG_OFFS] = (vec_t)hypot(re, im);
			vars[PHASE_OFFS] = cplx_phase(re, im);
			vars[NEG_FREQ_OFFS] = 0;
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[r][ftc] = vars[MAG_OFFS] * (vec_t)cos(vars[PHASE_OFFS]);
			rvmft->cell[r][ftc+1] = vars[MAG_OFFS] * (vec_t)sin(vars[PHASE_OFFS]);
			re = rvmft->cell[ftr][ftc];
			im = rvmft->cell[ftr][ftc+1];
			vars[MAG_OFFS] = (vec_t)hypot(re, im);
			vars[PHASE_OFFS] = cplx_phase(re, im);
			vars[NEG_FREQ_OFFS] = 1;
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[ftr][ftc] = vars[MAG_OFFS] * (vec_t)cos(vars[PHASE_OFFS]);
			rvmft->cell[ftr][ftc+1] = vars[MAG_OFFS] * (vec_t)sin(vars[PHASE_OFFS]);
		}
	}
	/* at highest frequency */
	if(r == (rvmft->rows >> 1))
	{
		nc = col_end - col_start + 1;
		ftc = col_start << 1;
		vars[ROW_OFFS] = (vec_t)r;
		for(;nc--;ftc += 2)
		{
			vars[COL_OFFS] = (vec_t)(ftc >> 1);
			re = rvmft->cell[r][ftc];
			im = rvmft->cell[r][ftc+1];
			vars[MAG_OFFS] = (vec_t)hypot(re, im);
			vars[PHASE_OFFS] = cplx_phase(re, im);
			vars[NEG_FREQ_OFFS] = 1;
			NSL_EXECUTE(trans->fn_main,vars);
			rvmft->cell[r][ftc] = vars[MAG_OFFS] * (vec_t)cos(vars[PHASE_OFFS]);
			rvmft->cell[r][ftc+1] = vars[MAG_OFFS] * (vec_t)sin(vars[PHASE_OFFS]);
		}
	}
}

int main()
{
	ipl_transform_p trans;
	vhandle_t lexer;
	matrix_p m = matAlloc(512,512);
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	lexer = lexParseFile("test.nsl","{}(),;");
	assert(lexer);
//	assert(lexNextIs(lexer,"transform"));
	lexNext(lexer);
	
	trans = iplLoadTransform(lexer);
	assert(trans);
	lexDestroy(lexer);

//	iplExecCartTrans(m,0,0,511,511,trans);
	iplExecTransform(m,0,0,511,511,trans);
	
//	iplExecVectorTransform(m,0,0,511,511,trans);

	iplFreeTransform(trans);
	save_image("out.bmp",m);
	memFree(m);
	memPrintStats();
	return 0;
}
